@@ -17,6 +17,8 @@
 #include <assert.h>
 #include <stddef.h>
 #include <poll.h>
+#include <limits.h>
+#include <locale.h>
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -25,6 +27,7 @@
 #include <sys/ioctl.h>
 #include <sys/wait.h>
 #include <sys/un.h>
+#include <sys/mman.h>
 
 #include <netinet/in.h>
 #include <netinet/udp.h>
@@ -70,9 +73,6 @@
 
 #define KBITS 0
 
-#define dispmode(mode)				\
-	(((mode) == KBITS) ? "kbps": "kBps")
-
 #define __noreturn	__attribute__((noreturn))
 #define __unused	__attribute__((unused))
 #define __printf(x, y)	__attribute__((format(printf, (x), (y))))
@@ -125,21 +125,21 @@ static inline char *skip_whitespace(char *str)
 	return str;
 }
 
-static inline unsigned long timeval_diff_msec(const struct timeval *end,
-					      const struct timeval *start)
+static inline unsigned long timespec_diff_msec(const struct timespec *end,
+					       const struct timespec *start)
 {
 	if (!start || !end)
 		return 0UL;
 
 	signed long secs = end->tv_sec - start->tv_sec;
-	signed long usecs = end->tv_usec - start->tv_usec;
+	signed long nsecs = end->tv_nsec - start->tv_nsec;
 
-	if(usecs < 0) {
-		usecs = 1000000 - usecs;
+	if(nsecs < 0) {
+		nsecs = 1000000000UL - nsecs;
 		secs -= 1;
 	}
 	if(secs >= 0)
-		return secs * 1000UL + usecs / 1000UL;
+		return secs * 1000UL + nsecs / 1000000UL;
 	else
 		return 0UL;
 }
