@@ -16,9 +16,7 @@ serv.c  - TCP/UDP port statistics module
 
 #include "dirs.h"
 #include "deskman.h"
-#include "isdntab.h"
 #include "fltdefs.h"
-#include "fltselect.h"
 #include "packet.h"
 #include "ipfrag.h"
 #include "ifaces.h"
@@ -27,13 +25,14 @@ serv.c  - TCP/UDP port statistics module
 #include "servname.h"
 #include "log.h"
 #include "timer.h"
-#include "promisc.h"
 #include "options.h"
-#include "instances.h"
 #include "packet.h"
 #include "logvars.h"
 #include "error.h"
-#include "bar.h"
+#include "counters.h"
+#include "rate.h"
+#include "capt.h"
+#include "timer.h"
 
 #define SCROLLUP 0
 #define SCROLLDOWN 1
@@ -41,29 +40,20 @@ serv.c  - TCP/UDP port statistics module
 #define LEFT 0
 #define RIGHT 1
 
-extern int exitloop;
-extern int daemonized;
-
-struct serv_spans {
-	int spanbr_in;
-	int spanbr_out;
-	int spanbr;
-};
-
 struct portlistent {
-	unsigned int port;
+	in_port_t port;
 	unsigned int protocol;
 	char servname[11];
 	unsigned int idx;
-	unsigned long long count;
-	unsigned long long bcount;
-	unsigned long long icount;
-	unsigned long long ibcount;
-	unsigned long long ocount;
-	unsigned long long obcount;
-	time_t starttime;
-	time_t proto_starttime;
-	struct serv_spans spans;
+	struct proto_counter serv_count;
+	struct proto_counter span;
+
+	struct timespec proto_starttime;
+
+	struct rate rate;
+	struct rate rate_in;
+	struct rate rate_out;
+
 	struct portlistent *prev_entry;
 	struct portlistent *next_entry;
 };
@@ -74,14 +64,14 @@ struct portlist {
 	struct portlistent *firstvisible;
 	struct portlistent *lastvisible;
 	struct portlistent *barptr;
-	int imaxy;
-	unsigned int baridx;
 	unsigned int count;
-	unsigned long bcount;
+
 	WINDOW *win;
 	PANEL *panel;
 	WINDOW *borderwin;
 	PANEL *borderpanel;
+	WINDOW *statwin;
+	PANEL *statpanel;
 };
 
 /*
@@ -95,67 +85,50 @@ static void rotate_serv_log(int s __unused)
 	signal(SIGUSR1, rotate_serv_log);
 }
 
-static void writeutslog(struct portlistent *list, unsigned long nsecs,
-			int units, FILE *fd)
+static void writeutslog(struct portlistent *list, unsigned long nsecs, FILE *fd)
 {
 	char atime[TIME_TARGET_MAX];
 	struct portlistent *ptmp = list;
-	float inrate, outrate, totalrate;
-	time_t now = time(NULL);
+	struct timespec now;
+
+	clock_gettime(CLOCK_MONOTONIC, &now);
 
 	genatime(time(NULL), atime);
 
 	fprintf(fd, "\n*** TCP/UDP traffic log, generated %s\n\n", atime);
 
 	while (ptmp != NULL) {
-		if (now - ptmp->proto_starttime < 5)
-			inrate = outrate = totalrate = -1.0;
-		else {
-			if (units == KBITS) {
-				inrate =
-				    (float) (ptmp->ibcount * 8 / 1000) /
-				    (float) (now - ptmp->proto_starttime);
-				outrate =
-				    (float) (ptmp->obcount * 8 / 1000) /
-				    (float) (now - ptmp->proto_starttime);
-				totalrate =
-				    (float) (ptmp->bcount * 8 / 1000) /
-				    (float) (now - ptmp->proto_starttime);
-			} else {
-				inrate =
-				    (float) (ptmp->obcount / 1024) /
-				    (float) (now - ptmp->proto_starttime);
-				outrate =
-				    (float) (ptmp->obcount / 1024) /
-				    (float) (now - ptmp->proto_starttime);
-				totalrate =
-				    (float) (ptmp->obcount / 1024) /
-				    (float) (now - ptmp->proto_starttime);
-			}
-		}
+		unsigned long secs = timespec_diff_msec(&now, &ptmp->proto_starttime) / 1000UL;
+		char bps_string[64];
 
 		if (ptmp->protocol == IPPROTO_TCP)
 			fprintf(fd, "TCP/%s: ", ptmp->servname);
 		else
 			fprintf(fd, "UDP/%s: ", ptmp->servname);
 
-		fprintf(fd, "%llu packets, %llu bytes total", ptmp->count,
-			ptmp->bcount);
+		fprintf(fd, "%llu packets, %llu bytes total",
+			ptmp->serv_count.proto_total.pc_packets,
+			ptmp->serv_count.proto_total.pc_bytes);
 
-		if (totalrate >= 0.0)
-			fprintf(fd, ", %.2f %s", totalrate, dispmode(units));
+		rate_print(ptmp->serv_count.proto_total.pc_bytes / secs,
+			   bps_string, sizeof(bps_string));
+		fprintf(fd, ", %s", bps_string);
 
-		fprintf(fd, "; %llu packets, %llu bytes incoming", ptmp->icount,
-			ptmp->ibcount);
+		fprintf(fd, "; %llu packets, %llu bytes incoming",
+			ptmp->serv_count.proto_in.pc_packets,
+			ptmp->serv_count.proto_in.pc_bytes);
 
-		if (inrate >= 0.0)
-			fprintf(fd, ", %.2f %s", inrate, dispmode(units));
+		rate_print(ptmp->serv_count.proto_in.pc_bytes / secs,
+			   bps_string, sizeof(bps_string));
+		fprintf(fd, ", %s", bps_string);
 
-		fprintf(fd, "; %llu packets, %llu bytes outgoing", ptmp->ocount,
-			ptmp->obcount);
+		fprintf(fd, "; %llu packets, %llu bytes outgoing",
+			ptmp->serv_count.proto_out.pc_packets,
+			ptmp->serv_count.proto_out.pc_bytes);
 
-		if (outrate >= 0.0)
-			fprintf(fd, ", %.2f %s", outrate, dispmode(units));
+		rate_print(ptmp->serv_count.proto_out.pc_bytes / secs,
+			   bps_string, sizeof(bps_string));
+		fprintf(fd, ", %s", bps_string);
 
 		fprintf(fd, "\n\n");
 		ptmp = ptmp->next_entry;
@@ -168,48 +141,74 @@ static void writeutslog(struct portlistent *list, unsigned long nsecs,
 static void initportlist(struct portlist *list)
 {
 	float screen_scale = ((float) COLS / 80 + 1) / 2;
-	int scratchx __unused;
 
 	list->head = list->tail = list->barptr = NULL;
 	list->firstvisible = list->lastvisible = NULL;
 	list->count = 0;
-	list->baridx = 0;
 
 	list->borderwin = newwin(LINES - 3, COLS, 1, 0);
 	list->borderpanel = new_panel(list->borderwin);
 	wattrset(list->borderwin, BOXATTR);
 	tx_box(list->borderwin, ACS_VLINE, ACS_HLINE);
 
-	wmove(list->borderwin, 0, 1 * screen_scale);
-	wprintw(list->borderwin, " Proto/Port ");
-	wmove(list->borderwin, 0, 22 * screen_scale);
-	wprintw(list->borderwin, " Pkts ");
-	wmove(list->borderwin, 0, 31 * screen_scale);
-	wprintw(list->borderwin, " Bytes ");
-	wmove(list->borderwin, 0, 40 * screen_scale);
-	wprintw(list->borderwin, " PktsTo ");
-	wmove(list->borderwin, 0, 49 * screen_scale);
-	wprintw(list->borderwin, " BytesTo ");
-	wmove(list->borderwin, 0, 58 * screen_scale);
-	wprintw(list->borderwin, " PktsFrom ");
-	wmove(list->borderwin, 0, 67 * screen_scale);
-	wprintw(list->borderwin, " BytesFrom ");
+	mvwprintw(list->borderwin, 0,  1 * screen_scale, " Proto/Port ");
+	mvwprintw(list->borderwin, 0, 22 * screen_scale, " Pkts ");
+	mvwprintw(list->borderwin, 0, 31 * screen_scale, " Bytes ");
+	mvwprintw(list->borderwin, 0, 40 * screen_scale, " PktsTo ");
+	mvwprintw(list->borderwin, 0, 49 * screen_scale, " BytesTo ");
+	mvwprintw(list->borderwin, 0, 58 * screen_scale, " PktsFrom ");
+	mvwprintw(list->borderwin, 0, 67 * screen_scale, " BytesFrom ");
 
 	list->win = newwin(LINES - 5, COLS - 2, 2, 1);
 	list->panel = new_panel(list->win);
-	getmaxyx(list->win, list->imaxy, scratchx);
+
+	list->statwin = newwin(1, COLS, LINES - 2, 0);
+	list->statpanel = new_panel(list->statwin);
+	scrollok(list->statwin, 0);
+	wattrset(list->statwin, IPSTATLABELATTR);
+	mvwprintw(list->statwin, 0, 0, "%*c", COLS, ' ');
 
 	tx_stdwinset(list->win);
 	wtimeout(list->win, -1);
 	wattrset(list->win, STDATTR);
 	tx_colorwin(list->win);
+
+	move(LINES - 1, 1);
+	scrollkeyhelp();
+	sortkeyhelp();
+	stdexitkeyhelp();
+
 	update_panels();
 	doupdate();
 }
 
+static void print_serv_rates(struct portlist *table)
+{
+	if (table->barptr == NULL) {
+		wattrset(table->statwin, IPSTATATTR);
+		mvwprintw(table->statwin, 0, 1, "No entries");
+	} else {
+		char buf[64];
+
+		wattrset(table->statwin, IPSTATLABELATTR);
+		mvwprintw(table->statwin, 0, 1, "Protocol data rates:");
+		mvwprintw(table->statwin, 0, 36, "total");
+		mvwprintw(table->statwin, 0, 57, "in");
+		mvwprintw(table->statwin, 0, 76, "out");
+
+		wattrset(table->statwin, IPSTATATTR);
+		rate_print(rate_get_average(&table->barptr->rate), buf, sizeof(buf));
+		mvwprintw(table->statwin, 0, 21, "%s", buf);
+		rate_print(rate_get_average(&table->barptr->rate_in), buf, sizeof(buf));
+		mvwprintw(table->statwin, 0, 42, "%s", buf);
+		rate_print(rate_get_average(&table->barptr->rate_out), buf, sizeof(buf));
+		mvwprintw(table->statwin, 0, 61, "%s", buf);
+	}
+}
+
 static struct portlistent *addtoportlist(struct portlist *list,
 					 unsigned int protocol,
-					 unsigned int port, int servnames)
+					 in_port_t port)
 {
 	struct portlistent *ptemp;
 
@@ -229,32 +228,36 @@ static struct portlistent *addtoportlist(struct portlist *list,
 
 	ptemp->protocol = protocol;
 	ptemp->port = port;	/* This is used in checks later. */
+	rate_alloc(&ptemp->rate, 5);
+	rate_alloc(&ptemp->rate_in, 5);
+	rate_alloc(&ptemp->rate_out, 5);
 
 	/*
 	 * Obtain appropriate service name
 	 */
 
-	servlook(servnames, htons(port), protocol, ptemp->servname, 10);
+	servlook(port, protocol, ptemp->servname, 10);
 
-	ptemp->count = ptemp->bcount = 0;
-	ptemp->icount = ptemp->ibcount = 0;
-	ptemp->ocount = ptemp->obcount = 0;
+	proto_counter_reset(&ptemp->serv_count);
+	proto_counter_reset(&ptemp->span);
 
 	list->count++;
 	ptemp->idx = list->count;
 
-	ptemp->proto_starttime = time(NULL);
+	clock_gettime(CLOCK_MONOTONIC, &ptemp->proto_starttime);
 
-	if (list->count <= LINES - 5)
+	if (list->count <= (unsigned) LINES - 5)
 		list->lastvisible = ptemp;
 
-	wmove(list->borderwin, LINES - 4, 1);
-	wprintw(list->borderwin, " %u entries ", list->count);
+	mvwprintw(list->borderwin, LINES - 4, 1, " %u entries ", list->count);
+
+	if (list->barptr == NULL)
+		list->barptr = ptemp;
 
 	return ptemp;
 }
 
-static int portinlist(struct porttab *table, unsigned int port)
+static int portinlist(struct porttab *table, in_port_t port)
 {
 	struct porttab *ptmp = table;
 
@@ -269,13 +272,13 @@ static int portinlist(struct porttab *table, unsigned int port)
 	return 0;
 }
 
-static int goodport(unsigned int port, struct porttab *table)
+static int goodport(in_port_t port, struct porttab *table)
 {
 	return ((port < 1024) || (portinlist(table, port)));
 }
 
 static struct portlistent *inportlist(struct portlist *list,
-				      unsigned int protocol, unsigned int port)
+				      unsigned int protocol, in_port_t port)
 {
 	struct portlistent *ptmp = list->head;
 
@@ -289,20 +292,19 @@ static struct portlistent *inportlist(struct portlist *list,
 	return NULL;
 }
 
-static void printportent(struct portlist *list, struct portlistent *entry,
-		  unsigned int idx)
+static void printportent(struct portlist *list, struct portlistent *entry)
 {
 	unsigned int target_row;
 	float screen_scale = ((float) COLS / 80 + 1) / 2;
 	int tcplabelattr;
 	int udplabelattr;
 	int highattr;
-	char sp_buf[10];
 
-	if ((entry->idx < idx) || (entry->idx > idx + (LINES - 6)))
+	if ((entry->idx < list->firstvisible->idx) ||
+	    (entry->idx > list->lastvisible->idx))
 		return;
 
-	target_row = entry->idx - idx;
+	target_row = entry->idx - list->firstvisible->idx;
 
 	if (entry == list->barptr) {
 		tcplabelattr = BARSTDATTR;
@@ -315,9 +317,8 @@ static void printportent(struct portlist *list, struct portlistent *entry,
 	}
 
 	wattrset(list->win, tcplabelattr);
-	sprintf(sp_buf, "%%%dc", COLS - 2);
 	scrollok(list->win, 0);
-	mvwprintw(list->win, target_row, 0, sp_buf, ' ');
+	mvwprintw(list->win, target_row, 0, "%*c", COLS - 2, ' ');
 	scrollok(list->win, 1);
 
 	wmove(list->win, target_row, 1);
@@ -332,61 +333,68 @@ static void printportent(struct portlist *list, struct portlistent *entry,
 	wprintw(list->win, "/%s          ", entry->servname);
 	wattrset(list->win, highattr);
 	wmove(list->win, target_row, 17 * screen_scale);
-	printlargenum(entry->count, list->win);
+	printlargenum(entry->serv_count.proto_total.pc_packets, list->win);
 	wmove(list->win, target_row, 27 * screen_scale);
-	printlargenum(entry->bcount, list->win);
+	printlargenum(entry->serv_count.proto_total.pc_bytes, list->win);
 	wmove(list->win, target_row, 37 * screen_scale);
-	printlargenum(entry->icount, list->win);
+	printlargenum(entry->serv_count.proto_in.pc_packets, list->win);
 	wmove(list->win, target_row, 47 * screen_scale);
-	printlargenum(entry->ibcount, list->win);
+	printlargenum(entry->serv_count.proto_in.pc_bytes, list->win);
 	wmove(list->win, target_row, 57 * screen_scale);
-	printlargenum(entry->ocount, list->win);
+	printlargenum(entry->serv_count.proto_out.pc_packets, list->win);
 	wmove(list->win, target_row, 67 * screen_scale);
-	printlargenum(entry->obcount, list->win);
+	printlargenum(entry->serv_count.proto_out.pc_bytes, list->win);
 }
 
 static void destroyportlist(struct portlist *list)
 {
 	struct portlistent *ptmp = list->head;
-	struct portlistent *ctmp = NULL;
-
-	if (list->head != NULL)
-		ctmp = list->head->next_entry;
 
 	while (ptmp != NULL) {
-		free(ptmp);
-		ptmp = ctmp;
+		struct portlistent *ctmp = ptmp->next_entry;
 
-		if (ctmp != NULL)
-			ctmp = ctmp->next_entry;
+		rate_destroy(&ptmp->rate_out);
+		rate_destroy(&ptmp->rate_in);
+		rate_destroy(&ptmp->rate);
+		free(ptmp);
+
+		ptmp = ctmp;
 	}
+
+	del_panel(list->panel);
+	delwin(list->win);
+	del_panel(list->borderpanel);
+	delwin(list->borderwin);
+	del_panel(list->statpanel);
+	delwin(list->statwin);
+
+	update_panels();
+	doupdate();
 }
 
 static void updateportent(struct portlist *list, unsigned int protocol,
-			  unsigned int sport, unsigned int dport, int br,
-			  unsigned int idx, struct porttab *ports,
-			  int servnames)
+			  in_port_t sport, in_port_t dport, int br,
+			  struct porttab *ports)
 {
 	struct portlistent *sport_listent = NULL;
 	struct portlistent *dport_listent = NULL;
+	enum {
+		PORT_INCOMING = 0,
+		PORT_OUTGOING
+	};
 
 	if (goodport(sport, ports)) {
 		sport_listent = inportlist(list, protocol, sport);
 
 		if (!sport_listent)
 			sport_listent =
-				addtoportlist(list, protocol, sport, servnames);
+				addtoportlist(list, protocol, sport);
 
 		if (sport_listent == NULL)
 			return;
 
-		sport_listent->count++;
-		sport_listent->bcount += br;
-		sport_listent->spans.spanbr += br;
-
-		sport_listent->obcount += br;
-		sport_listent->spans.spanbr_out += br;
-		sport_listent->ocount++;
+		proto_counter_update(&sport_listent->serv_count, PORT_OUTGOING, br);
+		proto_counter_update(&sport_listent->span, PORT_OUTGOING, br);
 	}
 
 	if (goodport(dport, ports)) {
@@ -394,26 +402,13 @@ static void updateportent(struct portlist *list, unsigned int protocol,
 
 		if (!dport_listent)
 			dport_listent =
-				addtoportlist(list, protocol, dport, servnames);
+				addtoportlist(list, protocol, dport);
 
 		if (dport_listent == NULL)
 			return;
 
-		if (dport_listent != sport_listent) {
-			dport_listent->count++;
-			dport_listent->bcount += br;
-			dport_listent->spans.spanbr += br;
-		}
-
-		dport_listent->ibcount += br;
-		dport_listent->spans.spanbr_in += br;
-		dport_listent->icount++;
-	}
-	if (sport_listent != NULL || dport_listent != NULL) {
-		if (sport_listent != NULL)
-			printportent(list, sport_listent, idx);
-		if (dport_listent != NULL && dport_listent != sport_listent)
-			printportent(list, dport_listent, idx);
+		proto_counter_update(&dport_listent->serv_count, PORT_INCOMING, br);
+		proto_counter_update(&dport_listent->span, PORT_INCOMING, br);
 	}
 }
 
@@ -474,22 +469,22 @@ static unsigned long long qp_getkey(struct portlistent *entry, int ch)
 		result = entry->port;
 		break;
 	case 'B':
-		result = entry->bcount;
+		result = entry->serv_count.proto_total.pc_bytes;
 		break;
 	case 'O':
-		result = entry->ibcount;
+		result = entry->serv_count.proto_in.pc_bytes;
 		break;
 	case 'M':
-		result = entry->obcount;
+		result = entry->serv_count.proto_out.pc_bytes;
 		break;
 	case 'P':
-		result = entry->count;
+		result = entry->serv_count.proto_total.pc_packets;
 		break;
 	case 'T':
-		result = entry->icount;
+		result = entry->serv_count.proto_in.pc_packets;
 		break;
 	case 'F':
-		result = entry->ocount;
+		result = entry->serv_count.proto_out.pc_packets;
 		break;
 	}
 
@@ -500,7 +495,7 @@ static unsigned long long qp_getkey(struct portlistent *entry, int ch)
  * Refresh TCP/UDP service screen.
  */
 
-static void refresh_serv_screen(struct portlist *table, int idx)
+static void refresh_serv_screen(struct portlist *table)
 {
 	struct portlistent *ptmp = table->firstvisible;
 
@@ -508,7 +503,7 @@ static void refresh_serv_screen(struct portlist *table, int idx)
 	tx_colorwin(table->win);
 
 	while ((ptmp != NULL) && (ptmp->prev_entry != table->lastvisible)) {
-		printportent(table, ptmp, idx);
+		printportent(table, ptmp);
 		ptmp = ptmp->next_entry;
 	}
 	update_panels();
@@ -626,11 +621,8 @@ static void quicksort_port_entries(struct portlist *table,
 	}
 }
 
-static void sortportents(struct portlist *list, unsigned int *idx, int command)
+static void sortportents(struct portlist *list, int command)
 {
-	struct portlistent *ptemp1;
-	int idxtmp;
-
 	if (!(list->head))
 		return;
 
@@ -643,72 +635,111 @@ static void sortportents(struct portlist *list, unsigned int *idx, int command)
 
 	quicksort_port_entries(list, list->head, list->tail, command);
 
-	ptemp1 = list->firstvisible = list->head;
-	*idx = 1;
-	idxtmp = 1;
-
-	while ((ptemp1) && (idxtmp <= LINES - 5)) {	/* printout */
-		printportent(list, ptemp1, *idx);
-		if (idxtmp <= LINES - 5)
-			list->lastvisible = ptemp1;
-		ptemp1 = ptemp1->next_entry;
-		idxtmp++;
+	list->firstvisible = list->head;
+	struct portlistent *ptmp = list->head;
+	while (ptmp && ((int)ptmp->idx <= getmaxy(list->win))) {
+		list->lastvisible = ptmp;
+		ptmp = ptmp->next_entry;
 	}
 }
 
-static void scrollservwin(struct portlist *table, int direction,
-			  unsigned int *idx)
+static void scrollservwin(struct portlist *table, int direction)
 {
-	char sp_buf[10];
-
-	sprintf(sp_buf, "%%%dc", COLS - 2);
 	wattrset(table->win, STDATTR);
 	if (direction == SCROLLUP) {
 		if (table->lastvisible != table->tail) {
-			wscrl(table->win, 1);
-			table->lastvisible = table->lastvisible->next_entry;
 			table->firstvisible = table->firstvisible->next_entry;
-			(*idx)++;
-			wmove(table->win, LINES - 6, 0);
+			table->lastvisible = table->lastvisible->next_entry;
+
+			wscrl(table->win, 1);
 			scrollok(table->win, 0);
-			wprintw(table->win, sp_buf, ' ');
+			mvwprintw(table->win, LINES - 6, 0, "%*c", COLS - 2, ' ');
 			scrollok(table->win, 1);
-			printportent(table, table->lastvisible, *idx);
+
+			printportent(table, table->lastvisible);
 		}
 	} else {
 		if (table->firstvisible != table->head) {
-			wscrl(table->win, -1);
-			table->lastvisible = table->lastvisible->prev_entry;
 			table->firstvisible = table->firstvisible->prev_entry;
-			(*idx)--;
-			wmove(table->win, 0, 0);
-			wprintw(table->win, sp_buf, ' ');
-			printportent(table, table->firstvisible, *idx);
+			table->lastvisible = table->lastvisible->prev_entry;
+
+			wscrl(table->win, -1);
+			mvwprintw(table->win, 0, 0, "%*c", COLS - 2, ' ');
+
+			printportent(table, table->firstvisible);
 		}
 	}
 }
 
-static void pageservwin(struct portlist *table, int direction,
-			unsigned int *idx)
+static void move_bar_one(struct portlist *table, int direction)
 {
-	int i = 1;
+	switch (direction) {
+	case SCROLLDOWN:
+		if (table->barptr->prev_entry == NULL)
+			break;
 
-	if (direction == SCROLLUP) {
-		while ((i <= LINES - 9) && (table->lastvisible != table->tail)) {
-			i++;
+		if (table->barptr == table->firstvisible)
+			scrollservwin(table, SCROLLDOWN);
+
+		table->barptr = table->barptr->prev_entry;
+		printportent(table, table->barptr->next_entry);	/* hide bar */
+		printportent(table, table->barptr);		/* show bar */
+
+		break;
+	case SCROLLUP:
+		if (table->barptr->next_entry == NULL)
+			break;
+
+		if (table->barptr == table->lastvisible)
+			scrollservwin(table, SCROLLUP);
+
+		table->barptr = table->barptr->next_entry;
+		printportent(table, table->barptr->prev_entry);	/* hide bar */
+		printportent(table, table->barptr);		/* show bar */
+
+		break;
+	}
+}
+
+static void move_bar_many(struct portlist *table, int direction, int lines)
+{
+	switch (direction) {
+	case SCROLLUP:
+		while (lines && (table->lastvisible != table->tail)) {
 			table->firstvisible = table->firstvisible->next_entry;
 			table->lastvisible = table->lastvisible->next_entry;
-			(*idx)++;
+			lines--;
 		}
-	} else {
-		while ((i <= LINES - 9) && (table->firstvisible != table->head)) {
-			i++;
+		if (lines == 0)
+			table->barptr = table->firstvisible;
+		else
+			table->barptr = table->lastvisible;
+		break;
+	case SCROLLDOWN:
+		while (lines && (table->firstvisible != table->head)) {
 			table->firstvisible = table->firstvisible->prev_entry;
 			table->lastvisible = table->lastvisible->prev_entry;
-			(*idx)--;
+			lines--;
 		}
+		table->barptr = table->firstvisible;
+		break;
 	}
-	refresh_serv_screen(table, *idx);
+	refresh_serv_screen(table);
+}
+
+static void move_bar(struct portlist *table, int direction, int lines)
+{
+	if (table->barptr == NULL)
+		return;
+	if (lines < 1)
+		return;
+	if (lines < 16)
+		while (lines--)
+			move_bar_one(table, direction);
+	else
+		move_bar_many(table, direction, lines);
+
+	print_serv_rates(table);
 }
 
 static void show_portsort_keywin(WINDOW ** win, PANEL ** panel)
@@ -744,139 +775,149 @@ static void show_portsort_keywin(WINDOW ** win, PANEL ** panel)
 	doupdate();
 }
 
-static void update_serv_rates(struct portlist *list, WINDOW * win, int actmode,
-		       int *cleared)
+static void update_serv_rates(struct portlist *list, unsigned long msecs)
 {
-	float inrate, outrate, totalrate;
-	time_t now = time(NULL);
+	/* update rates of all portlistents */
+	for (struct portlistent *ple = list->head; ple != NULL; ple = ple->next_entry) {
+		rate_add_rate(&ple->rate, ple->span.proto_total.pc_bytes, msecs);
+		rate_add_rate(&ple->rate_in, ple->span.proto_in.pc_bytes, msecs);
+		rate_add_rate(&ple->rate_out, ple->span.proto_out.pc_bytes, msecs);
 
-	if (actmode == KBITS) {
-		inrate =
-		    (float) (list->barptr->spans.spanbr_in * 8 / 1000) /
-		    (float) (now - list->barptr->starttime);
-		outrate =
-		    (float) (list->barptr->spans.spanbr_out * 8 / 1000) /
-		    (float) (now - list->barptr->starttime);
-		totalrate =
-		    (float) (list->barptr->spans.spanbr * 8 / 1000) /
-		    (float) (now - list->barptr->starttime);
-	} else {
-		inrate =
-		    (float) (list->barptr->spans.spanbr_in / 1024) /
-		    (float) (now - list->barptr->starttime);
-		outrate =
-		    (float) (list->barptr->spans.spanbr_out / 1024) /
-		    (float) (now - list->barptr->starttime);
-		totalrate =
-		    (float) (list->barptr->spans.spanbr / 1024) / (float) (now -
-									   list->
-									   barptr->
-									   starttime);
+		proto_counter_reset(&ple->span);
+	}
+}
+
+static void serv_process_key(struct portlist *table, int ch)
+{
+	static WINDOW *sortwin;
+	static PANEL *sortpanel;
+
+	static int keymode = 0;
+
+	if (keymode == 0) {
+		switch (ch) {
+		case KEY_UP:
+			move_bar(table, SCROLLDOWN, 1);
+			break;
+		case KEY_DOWN:
+			move_bar(table, SCROLLUP, 1);
+			break;
+		case KEY_PPAGE:
+		case '-':
+			move_bar(table, SCROLLDOWN, LINES - 5);
+			break;
+		case KEY_NPAGE:
+		case ' ':
+			move_bar(table, SCROLLUP, LINES - 5);
+			break;
+		case KEY_HOME:
+			move_bar(table, SCROLLDOWN, INT_MAX);
+			break;
+		case KEY_END:
+			move_bar(table, SCROLLUP, INT_MAX);
+			break;
+		case 12:
+		case 'l':
+		case 'L':
+			tx_refresh_screen();
+			break;
+		case 's':
+		case 'S':
+			show_portsort_keywin(&sortwin, &sortpanel);
+			keymode = 1;
+			break;
+		case 'q':
+		case 'Q':
+		case 'x':
+		case 'X':
+		case 27:
+		case 24:
+			exitloop = 1;
+		}
+	} else if (keymode == 1) {
+		del_panel(sortpanel);
+		delwin(sortwin);
+		sortportents(table, ch);
+		keymode = 0;
+		refresh_serv_screen(table);
+		table->barptr = table->firstvisible;
+		print_serv_rates(table);
+		update_panels();
+		doupdate();
+	}
+}
+
+static void serv_process_packet(struct portlist *table, struct pkt_hdr *pkt,
+				struct porttab *ports)
+{
+	unsigned int tot_br;
+	in_port_t sport = 0;
+	in_port_t dport = 0;
+
+	int pkt_result = packet_process(pkt, &tot_br, &sport, &dport,
+					MATCH_OPPOSITE_USECONFIG,
+					options.v6inv4asv6);
+
+	if (pkt_result != PACKET_OK)
+		return;
+
+	unsigned short iplen;
+	switch (pkt->pkt_protocol) {
+	case ETH_P_IP:
+		iplen =	ntohs(pkt->iphdr->tot_len);
+		break;
+	case ETH_P_IPV6:
+		iplen = ntohs(pkt->ip6_hdr->ip6_plen) + 40;
+		break;
+	default:
+		/* unknown link protocol */
+		return;
 	}
 
-	wattrset(win, IPSTATLABELATTR);
-	mvwprintw(win, 0, 1,
-		  "Protocol data rates (%s):              in             out            total",
-		  dispmode(actmode));
-	wattrset(win, IPSTATATTR);
-	mvwprintw(win, 0, 31, "%10.2f", inrate);
-	mvwprintw(win, 0, 46, "%10.2f", outrate);
-	mvwprintw(win, 0, 61, "%10.2f", totalrate);
-
-	memset(&(list->barptr->spans), 0, sizeof(struct serv_spans));
-	list->barptr->starttime = time(NULL);
-	*cleared = 0;
+	__u8 ip_protocol = pkt_ip_protocol(pkt);
+	switch (ip_protocol) {
+	case IPPROTO_TCP:
+	case IPPROTO_UDP:
+		updateportent(table, ip_protocol, sport, dport, iplen, ports);
+		break;
+	default:
+		/* unknown L4 protocol */
+		return;
+	}
 }
 
 /*
  * The TCP/UDP service monitor
  */
 
-void servmon(char *ifname, struct porttab *ports, const struct OPTIONS *options,
-	     time_t facilitytime, struct filterstate *ofilter)
+void servmon(char *ifname, time_t facilitytime)
 {
-	int logging = options->logging;
-	int pkt_result;
-
-	int keymode = 0;
-
-	unsigned int idx = 1;
-
-	unsigned int sport = 0;
-	unsigned int dport = 0;
-
-	struct timeval tv;
-	time_t starttime, startlog, timeint;
-	time_t now;
-	unsigned long long unow;
-	time_t updtime = 0;
-	unsigned long long updtime_usec = 0;
-
-	unsigned int tot_br;
+	int logging = options.logging;
 
 	int ch;
 
 	struct portlist list;
-	struct portlistent *serv_tmp;
-	int statcleared = 0;
 
 	FILE *logfile = NULL;
 
-	struct promisc_states *promisc_list;
+	struct capt capt;
 
-	WINDOW *sortwin;
-	PANEL *sortpanel;
+	struct porttab *ports;
 
-	WINDOW *statwin;
-	PANEL *statpanel;
-
-	char sp_buf[10];
-
-	const int statx = 1;
-
-	int fd;
-
-	/*
-	 * Mark this facility
-	 */
-
-	if (!facility_active(TCPUDPIDFILE, ifname))
-		mark_facility(TCPUDPIDFILE, "TCP/UDP monitor", ifname);
-	else {
-		write_error("TCP/UDP monitor already running on %s", ifname);
-		return;
-	}
+	struct pkt_hdr pkt;
 
 	if (!dev_up(ifname)) {
 		err_iface_down();
-		unmark_facility(TCPUDPIDFILE, ifname);
 		return;
 	}
 
-	if ((first_active_facility()) && (options->promisc)) {
-		init_promisc_list(&promisc_list);
-		save_promisc_list(promisc_list);
-		srpromisc(1, promisc_list);
-		destroy_promisc_list(&promisc_list);
-	}
-
-	adjust_instance_count(PROCCOUNTFILE, 1);
-
 	initportlist(&list);
-	statwin = newwin(1, COLS, LINES - 2, 0);
-	statpanel = new_panel(statwin);
-	scrollok(statwin, 0);
-	wattrset(statwin, IPSTATLABELATTR);
-	sprintf(sp_buf, "%%%dc", COLS);
-	mvwprintw(statwin, 0, 0, sp_buf, ' ');
+	loadaddports(&ports);
 
-	move(LINES - 1, 1);
-	scrollkeyhelp();
-	sortkeyhelp();
-	stdexitkeyhelp();
-
-	if (options->servnames)
-		setservent(1);
+	if (capt_init(&capt, ifname) == -1) {
+		write_error("Unable to initialize packet capture interface");
+		goto err;
+	}
 
 	if (logging) {
 		if (strcmp(current_logfile, "") == 0) {
@@ -902,284 +943,97 @@ void servmon(char *ifname, struct porttab *ports, const struct OPTIONS *options,
 			 "******** TCP/UDP service monitor started ********");
 	}
 
-	//isdnfd = -1;
+	if (options.servnames)
+		setservent(1);
+
+	packet_init(&pkt);
+
 	exitloop = 0;
-	gettimeofday(&tv, NULL);
-	starttime = startlog = timeint = tv.tv_sec;
 
-	wattrset(statwin, IPSTATATTR);
-	mvwprintw(statwin, 0, 1, "No entries");
-	update_panels();
-	doupdate();
+	struct timespec now;
+	clock_gettime(CLOCK_MONOTONIC, &now);
+	struct timespec last_time = now;
+	struct timespec next_screen_update = { 0 };
+	time_t starttime = now.tv_sec;
+	time_t endtime = INT_MAX;
+	if (facilitytime != 0)
+		endtime = now.tv_sec + facilitytime * 60;
 
-	fd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
-	if(fd == -1) {
-		write_error("Unable to obtain monitoring socket");
-		goto err;
-	}
-	if(dev_bind_ifname(fd, ifname) == -1) {
-		write_error("Unable to bind interface on the socket");
-		goto err_close;
-	}
-
-	PACKET_INIT(pkt);
+	time_t log_next = INT_MAX;
+	if (logging)
+		log_next = now.tv_sec + options.logspan;
 
 	while (!exitloop) {
-		gettimeofday(&tv, NULL);
-		now = tv.tv_sec;
-		unow = tv.tv_sec * 1000000ULL + tv.tv_usec;
+		clock_gettime(CLOCK_MONOTONIC, &now);
 
-		if (now - timeint >= 5) {
-			printelapsedtime(starttime, now, LINES - 4, 20,
-					 list.borderwin);
-			timeint = now;
-		}
-		if (logging) {
-			check_rotate_flag(&logfile);
-			if ((now - startlog) >= options->logspan) {
-				writeutslog(list.head, now - starttime,
-					    options->actmode, logfile);
-				startlog = now;
+		if (now.tv_sec > last_time.tv_sec) {
+			unsigned long rate_msecs = timespec_diff_msec(&now, &last_time);
+			/* update all portlistent rates ... */
+			update_serv_rates(&list, rate_msecs);
+			/* ... and print the current one */
+			print_serv_rates(&list);
+
+			printelapsedtime(now.tv_sec - starttime, 20, list.borderwin);
+
+			print_packet_drops(capt_get_dropped(&capt), list.borderwin, 49);
+
+			if (now.tv_sec > endtime)
+				exitloop = 1;
+
+			if (logging && (now.tv_sec > log_next)) {
+				check_rotate_flag(&logfile);
+				writeutslog(list.head, now.tv_sec - starttime,
+					    logfile);
+				log_next = now.tv_sec + options.logspan;
 			}
+
+			last_time = now;
 		}
 
-		if (list.barptr != NULL) {
-			if ((now - list.barptr->starttime) >= 5) {
-				update_serv_rates(&list, statwin,
-						  options->actmode,
-						  &statcleared);
-			}
-		}
+		if (time_after(&now, &next_screen_update)) {
+			refresh_serv_screen(&list);
 
-		if (((options->updrate != 0)
-		     && (now - updtime >= options->updrate))
-		    || ((options->updrate == 0)
-			&& (unow - updtime_usec >= DEFAULT_UPDATE_DELAY))) {
 			update_panels();
 			doupdate();
-			updtime = now;
-			updtime_usec = unow;
+
+			set_next_screen_update(&next_screen_update, &now);
 		}
 
-		if ((facilitytime != 0)
-		    && (((now - starttime) / 60) >= facilitytime))
-			exitloop = 1;
-
-		if (packet_get(fd, &pkt, &ch, list.win) == -1) {
+		if (capt_get_packet(&capt, &pkt, &ch, list.win) == -1) {
 			write_error("Packet receive failed");
 			exitloop = 1;
 			break;
 		}
 
-		if (ch == ERR)
-			goto no_key_ready;
+		if (ch != ERR)
+			serv_process_key(&list, ch);
 
-		if (keymode == 0) {
-			switch (ch) {
-			case KEY_UP:
-				if (!list.barptr
-				    || !list.barptr->prev_entry)
-					break;
-
-				serv_tmp = list.barptr;
-				set_barptr((void *) &(list.barptr),
-					   list.barptr->prev_entry,
-					   &(list.barptr->prev_entry->starttime),
-					   &(list.barptr->prev_entry->spans),
-					   sizeof(struct serv_spans),
-					   statwin, &statcleared, statx);
-				printportent(&list, serv_tmp, idx);
-
-				if (list.baridx == 1)
-					scrollservwin(&list, SCROLLDOWN, &idx);
-				else
-					list.baridx--;
-
-				printportent(&list, list.barptr, idx);
-				break;
-			case KEY_DOWN:
-				if (!list.barptr
-				    || !list.barptr->next_entry)
-					break;
-
-				serv_tmp = list.barptr;
-				set_barptr((void *) &(list.barptr),
-					   list.barptr->next_entry,
-					   &(list.barptr->next_entry->starttime),
-					   &(list.barptr->next_entry->spans),
-					   sizeof(struct serv_spans),
-					   statwin, &statcleared, statx);
-				printportent(&list,serv_tmp, idx);
-
-				if (list.baridx == list.imaxy)
-					scrollservwin(&list, SCROLLUP, &idx);
-				else
-					list.baridx++;
-
-				printportent(&list, list.barptr, idx);
-				break;
-			case KEY_PPAGE:
-			case '-':
-				if (!list.barptr)
-					break;
-
-				pageservwin(&list, SCROLLDOWN, &idx);
-
-				set_barptr((void *) &(list.barptr),
-					   list.lastvisible,
-					   &(list.lastvisible->starttime),
-					   &(list.lastvisible->spans),
-					   sizeof(struct serv_spans),
-					   statwin, &statcleared, statx);
-				list.baridx = list.lastvisible->idx - idx + 1;
-
-				refresh_serv_screen(&list, idx);
-				break;
-			case KEY_NPAGE:
-			case ' ':
-				if (!list.barptr)
-					break;
-
-				pageservwin(&list, SCROLLUP, &idx);
-
-				set_barptr((void *) &(list.barptr),
-					   list.firstvisible,
-					   &(list.firstvisible->starttime),
-					   &(list.firstvisible->spans),
-					   sizeof(struct serv_spans),
-					   statwin, &statcleared, statx);
-				list.baridx = 1;
-
-				refresh_serv_screen(&list, idx);
-
-				break;
-			case 12:
-			case 'l':
-			case 'L':
-				tx_refresh_screen();
-				break;
-			case 's':
-			case 'S':
-				show_portsort_keywin(&sortwin,
-						     &sortpanel);
-				keymode = 1;
-				break;
-			case 'q':
-			case 'Q':
-			case 'x':
-			case 'X':
-			case 27:
-			case 24:
-				exitloop = 1;
-			}
-		} else if (keymode == 1) {
-			del_panel(sortpanel);
-			delwin(sortwin);
-			sortportents(&list, &idx, ch);
-			keymode = 0;
-			if (list.barptr != NULL) {
-				set_barptr((void *) &(list.barptr),
-					   list.firstvisible,
-					   &(list.firstvisible->
-					     starttime),
-					   &(list.firstvisible->spans),
-					   sizeof(struct serv_spans),
-					   statwin, &statcleared,
-					   statx);
-				list.baridx = 1;
-			}
-			refresh_serv_screen(&list, idx);
-			update_panels();
-			doupdate();
-		}
-	no_key_ready:
-
-		if (pkt.pkt_len <= 0)
-			continue;
-
-		unsigned short ipproto;
-		unsigned short iplen;
-
-		pkt_result =
-			packet_process(&pkt, &tot_br, &sport, &dport,
-				      ofilter,
-				      MATCH_OPPOSITE_USECONFIG,
-				      options->v6inv4asv6);
-
-		if (pkt_result != PACKET_OK)
-			continue;
-
-		switch (pkt.pkt_protocol) {
-		case ETH_P_IP:
-			ipproto = ((struct iphdr *) pkt.pkt_payload)->protocol;
-			iplen =	ntohs(((struct iphdr *) pkt.pkt_payload)->tot_len);
-			break;
-		case ETH_P_IPV6:
-			ipproto = ((struct ip6_hdr *) pkt.pkt_payload)->ip6_nxt;	/* FIXME: extension headers ??? */
-			iplen = ntohs(((struct ip6_hdr *) pkt.pkt_payload)->ip6_plen) + 40;
-			break;
-		default:
-			/* unknown link protocol */
-			continue;
-		}
-		switch (ipproto) {
-		case IPPROTO_TCP:
-		case IPPROTO_UDP:
-			updateportent(&list, ipproto, ntohs(sport),
-				      ntohs(dport), iplen, idx, ports,
-				      options->servnames);
-			break;
-		default:
-			/* unknown L4 protocol */
-			continue;
-		}
-		if ((list.barptr == NULL) && (list.head != NULL)) {
-			set_barptr((void *) &(list.barptr), list.head,
-				   &(list.head->starttime),
-				   &(list.head->spans),
-				   sizeof(struct serv_spans), statwin,
-				   &statcleared, statx);
-			list.baridx = 1;
+		if (pkt.pkt_len > 0) {
+			serv_process_packet(&list, &pkt, ports);
+			capt_put_packet(&capt, &pkt);
 		}
 	}
+	packet_destroy(&pkt);
+
+	if (options.servnames)
+		endservent();
 
-err_close:
-	close(fd);
-err:
 	if (logging) {
 		signal(SIGUSR1, SIG_DFL);
-		writeutslog(list.head, time(NULL) - starttime,
-			    options->actmode, logfile);
+		writeutslog(list.head, time(NULL) - starttime, logfile);
 		writelog(logging, logfile,
 			 "******** TCP/UDP service monitor stopped ********");
 		fclose(logfile);
 	}
-	if (options->servnames)
-		endservent();
-
-	if ((options->promisc) && (is_last_instance())) {
-		load_promisc_list(&promisc_list);
-		srpromisc(0, promisc_list);
-		destroy_promisc_list(&promisc_list);
-	}
-
-	adjust_instance_count(PROCCOUNTFILE, -1);
-
-	del_panel(list.panel);
-	delwin(list.win);
-	del_panel(list.borderpanel);
-	delwin(list.borderwin);
-	del_panel(statpanel);
-	delwin(statwin);
-	unmark_facility(TCPUDPIDFILE, ifname);
-	update_panels();
-	doupdate();
-	destroyportlist(&list);
-	pkt_cleanup();
 	strcpy(current_logfile, "");
+
+	capt_destroy(&capt);
+err:
+	destroyporttab(ports);
+	destroyportlist(&list);
 }
 
-static void portdlg(unsigned int *port_min, unsigned int *port_max,
+static void portdlg(in_port_t *port_min, in_port_t *port_max,
 		    int *aborted)
 {
 	WINDOW *bw;
@@ -1305,7 +1159,7 @@ static int dup_portentry(struct porttab *table, unsigned int min,
 
 void addmoreports(struct porttab **table)
 {
-	unsigned int port_min, port_max;
+	in_port_t port_min = 0, port_max = 0;
 	int aborted;
 	struct porttab *ptmp;
 
@@ -1380,7 +1234,6 @@ void loadaddports(struct porttab **table)
 static void operate_portselect(struct porttab **table, struct porttab **node,
 			       int *aborted)
 {
-	int ch = 0;
 	struct scroll_list list;
 	char listtext[20];
 
@@ -1398,7 +1251,7 @@ static void operate_portselect(struct porttab **table, struct porttab **node,
 	}
 
 	tx_show_listbox(&list);
-	tx_operate_listbox(&list, &ch, aborted);
+	tx_operate_listbox(&list, aborted);
 
 	if (!(*aborted))
 		*node = (struct porttab *) list.textptr->nodeptr;
@@ -1439,11 +1292,11 @@ static void delport(struct porttab **table, struct porttab *ptmp)
 void removeaport(struct porttab **table)
 {
 	int aborted;
-	struct porttab *ptmp;
+	struct porttab *ptmp = NULL;
 
 	selectport(table, &ptmp, &aborted);
 
-	if (!aborted) {
+	if (!aborted && ptmp) {
 		delport(table, ptmp);
 		saveportlist(*table);
 	}
@@ -1451,17 +1304,10 @@ void removeaport(struct porttab **table)
 
 void destroyporttab(struct porttab *table)
 {
-	struct porttab *ptemp = table;
-	struct porttab *ctemp = NULL;
+	while (table != NULL) {
+		struct porttab *ptemp = table->next_entry;
 
-	if (ptemp != NULL)
-		ctemp = ptemp->next_entry;
-
-	while (ptemp != NULL) {
-		free(ptemp);
-		ptemp = ctemp;
-
-		if (ctemp != NULL)
-			ctemp = ctemp->next_entry;
+		free(table);
+		table = ptemp;
 	}
 }
